#include "JacQL_Base.h"
#include <dirent.h>
#include <errno.h>

void JacQL_Base::setDependency(dependency d, unsigned int d_id)
{
    dependencies[d] = d_id;
    saveRequest();
}

int JacQL_Base::compare(JacQL_Base *const& first, JacQL_Base *const& second)
{
    if(first->getId() < second->getId())
        return -1;
    else if(first->getId() == second->getId())
        return 0;
    return 1;
}

int JacQL_Base::compareID(const unsigned int& first, JacQL_Base *const& second)
{
    if(first < second->getId())
        return -1;
    else if(first == second->getId())
        return 0;
    return 1;
}

int JacQL_Base::compareUI(const unsigned int& first, const unsigned int& second)
{
    if(first < second)
        return -1;
    else if(first == second)
        return 0;
    return 1;
}

bool JacQL_Base::applyConstraints(JacQL_Base *const& element)
{
    const Constrainer& constraints = element->getConstraints();
    for(int i=0; i<abs(constraints.num); i++)
    {
        if(element->getDependency(constraints.conTable[i]) != constraints.conId[i])
        {
            if(constraints.num > 0)
                return false;
        }
        else if(constraints.num < 0)
            return false;
    }
    return true;
}

bool JacQL_Base::applyQuery(JacQL_Base *const& element)
{
    const QueryType& queryType = element->getQueryType();
    const int& queryId = element->getQueryId();
    const void *queryElement = element->getQueryElement();
    const void *dataElement = element->getData(queryId);
    const int& queryComparison = element->getQueryComparison();
    int compValue;

    if(queryType == QUERY_INVALID)
        throw JacQLException_Base("Fatal JacQLException: Invalid query type!");
    if(queryType == QUERY_STRING || queryType == QUERY_UNICODE) //String or Unicode
    {
        int strCompare;
        if(queryComparison > 0x06) //CONTAINS and !CONTAINS
        {
            void *temp;
            if(queryType == QUERY_STRING)
                temp = strstr((const char *)dataElement, (const char *)queryElement);
            else
                temp = wcsstr((const wchar_t *)dataElement, (const wchar_t *)queryElement);
            compValue = (temp == NULL)?0x10:0x08;
        }
        else //Standard comparisons
        {
            if(queryType == QUERY_STRING) //Standard string
                strCompare = strcmp((const char *)dataElement, (const char *)queryElement);
            else //Unicode string
                strCompare = wcscmp((const wchar_t *)dataElement, (const wchar_t *)queryElement);

            if(strCompare < 0)
                compValue = 0x2;
            else if(strCompare > 0)
                compValue = 0x4;
            else
                compValue = 0x1;
        }
    }
    else if(queryType == QUERY_NUMBER) //Long
    {
        long *data = (long *)(dataElement);
        long *query = (long *)(queryElement);
        if(*data == *query)
            compValue = 0x1;
        else if(*data < *query)
            compValue = 0x2;
        else
            compValue = 0x4;
    }
    else if(queryType == QUERY_FLOAT) //Double
    {
        double *data = (double *)(dataElement);
        double *query = (double *)(queryElement);
        if(*data == *query)
            compValue = 0x1;
        else if(*data < *query)
            compValue = 0x2;
        else
            compValue = 0x4;
    }
    else//Boolean
    {
        bool *data = (bool *)(dataElement);
        bool *query = (bool *)(queryElement);
        if(*data == *query)
            compValue = 0x1;
        else
            compValue = 0x6;
    }
    return (queryComparison & compValue);
}

int JacQL_Base::applyCompareQuery(JacQL_Base *const& first, JacQL_Base *const& second)
{
    const QueryType& queryType = first->getCompareQueryType();
    const int& queryId = first->getCompareQueryId();
    const void *firstElement = first->getData(queryId);
    const void *secondElement = second->getData(queryId);
    const bool compareQueryDesc = first->getCompareQueryDesc();
    int compValue;

    if(queryType == QUERY_INVALID)
        throw JacQLException_Base("Fatal JacQLException: Invalid query type!");
    if(queryType == QUERY_STRING || queryType == QUERY_UNICODE) //String or Unicode
    {
        if(queryType == QUERY_STRING) //Standard string
            compValue = strcmp((const char *)firstElement, (const char *)secondElement);
        else //Unicode string
            compValue = wcscmp((const wchar_t *)firstElement, (const wchar_t *)secondElement);
    }
    else if(queryType == QUERY_NUMBER) //Long
    {
        long *f = (long *)(firstElement);
        long *s = (long *)(secondElement);
        if(*f == *s)
            compValue = 0;
        else if(*f < *s)
            compValue = -1;
        else
            compValue = 1;
    }
    else if(queryType == QUERY_FLOAT) //Double
    {
        double *f = (double *)(firstElement);
        double *s = (double *)(secondElement);
        if(*f == *s)
            compValue = 0;
        else if(*f < *s)
            compValue = -1;
        else
            compValue = 1;
    }
    else//Boolean
    {
        bool *f = (bool *)(firstElement);
        bool *s = (bool *)(secondElement);
        if(*f == *s)
            compValue = 0;
        else if(*f == true)
            compValue = -1;
        else
            compValue = 1;
    }
    return (compValue*((compareQueryDesc)?-1:1));
}

void JacQL_Base::deleteImageFile(const char *directory)
{
    char fileLocation[strlen(directory)+15];
    sprintf(fileLocation, "%s%u.bmp", directory, id);
    remove(fileLocation);
}

void JacQL_Base::clearUpDir(const char *path, const bool fullClear)
{
    const int PATH_BUFFER_SIZE = 2056;
    char pathBuffer[PATH_BUFFER_SIZE];
    const char *filePrefix;
    if(fullClear)
        strcpy(pathBuffer, path);
    else
    {
        filePrefix = strrchr(path, '/');
        if(filePrefix == NULL)
            filePrefix = path;
        else
            filePrefix++; //Not on the slash but the preceding character
        int counter;
        for(counter = 0; counter < filePrefix-path; counter++)
            pathBuffer[counter] = path[counter];
        pathBuffer[counter] = '\0';
    }
    DIR *direct = opendir(pathBuffer);
    if(direct == NULL)
        throw JacQLException_Base("Fatal JacQL_Exception: In clearUpDir(), opendir() returned an error. Wrong path?");
    while(true)
    {
        struct dirent *entry = readdir(direct);
        if(entry == NULL)
            break;
        if((entry->d_name)[0] != '.' && (fullClear || strstr(entry->d_name, filePrefix) == entry->d_name))
        {
            char temp[PATH_BUFFER_SIZE];
            sprintf(temp, "%s%s", pathBuffer, entry->d_name);
            if(remove(temp) == -1)
                throw JacQLException_Base("Fatal JacQL_Exception: In clearUpDir(), could not remove file! File might be open or user has no access right!");
        }
    }
    closedir(direct);
}
