#ifndef JACQLNODE_H_INCLUDED
#define JACQLNODE_H_INCLUDED

/** JacQLNode - JacQL
  *
  *  By JackTex, 2016. All copyright reserved.
  *
  *  Node used for most data structures in JacQL. Interior is in public interface as many data structures
  *  manipulate nodes directly. Core element of JacQL. Do not change unless if you know what you are doing!
  **/

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * Functions dependent on class T * * * ////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

template<class T> inline
void delete_or_nop (T  const&)    {/* NOP */}

template<class T> inline
void delete_or_nop (T* const& p)  {delete p;}


template<class T> inline
void invalidate_or_nop (T* const&)    {/* NOP */}

template<class T> inline
void invalidate_or_nop (T** const& p)  {*p=NULL;}

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * JacQLNode * * * /////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

template <class T>
class JacQLNode
{
    public:
        T val;
        int balance;
        int quickHeight; //To reduce the need and cost of recursive calls in AVL tree. If -1: Currently unknown
        unsigned int relationCount; //Counter for relations
        JacQLNode *left, *right, *parent; //For tree structures
        JacQLNode *before, *after; //For list structures. WARNING: Do not use for iteration! Use next() prev() instead

        JacQLNode(T k, JacQLNode *p) : val(k), balance(0), quickHeight(0), relationCount(1U),
                                       left(NULL), right(NULL), parent(p), before(NULL), after(NULL) {}

        virtual ~JacQLNode() {
            if(relationCount>0)
                delete_or_nop(val);
            delete left;
            delete right;
            if(before!=NULL)
                before->after = after;
            if(after!=NULL)
                after->before = before;
        }

        JacQLNode *next(){return (after==NULL)?NULL:after->getAddress(this, true);}
        JacQLNode *prev(){return (before==NULL)?NULL:before->getAddress(this, false);}

    protected:
        JacQLNode() : relationCount(1U), left(NULL), right(NULL), parent(NULL), before(NULL), after(NULL) {invalidate_or_nop(&val);}
        virtual JacQLNode *getAddress(JacQLNode<T> *requester, bool forwardGoing){return this;} //This hack is needed
            //because of the selectionHeap which is a very special kind of JacQLNode (see JacQL_AVL_Tree.h)

};


#endif // JACQLNODE_H_INCLUDED
