#ifndef JACQL_TEMPLATE_H
#define JACQL_TEMPLATE_H

/** JacQL Template - JacQL
  *
  *  By JackTex, 2016. All copyright reserved.
  *
  *  This template contains all abstract and static methods which should be different for each table.
  *  In particular, all access to the element tree as well as saving/loading adding/deleting entries
  *  is managed through this template. Only more generalisable properties are covered by JacQL_Base.h
  *  Moreover, note that all DATAITERATOR and SELITERATOR methods can an alternative element tree
  *  as an argument. This property is particular useful for iterating through first and second tier
  *  relation AVL trees. For instance a query can be performed by simply calling the static method
  *  and giving the according AVL tree to be modified.
  *
  *  JacQL_Template<T> has the public interface:
  *     virtual void setDependency(dependency, unsigned int); //Sets the dependency to another entry
  *     virtual T& operator=(const T&)=0; //Assign operator for T
  *
  *     //dataIterator and selIterator are wrappers around the treeIterator such that dereferencing
  *     //gives a pointer to T rather than JacQL_Base
  *
  *     static void setup(dependency, bool=false); //Sets the dependencyTruthTable to all false and sets the own Entry and if it is a weak Table (weak tables do not have
  *            //any dependencies from other tables pointing to it and some database overhead can be reduced by explicitly declaring this circumstance -for instance: Associations-)
  *     static void loadAll(const char *, dependency) throw (InvalidFileException); //Loads all values in the given path
  *     static void saveAll(const char *, dependency) throw (InvalidFileException); //Saves all values in the given path
  *     static void deleteEntriesWithDependency(dependency, unsigned int); //Deletes all entries with dependency to the stated id and calls deletion recursively
  *     static void updateEntriesWithDependency(dependency, unsigned int, unsigned int); //Changes the dependencies of all entries with this one
  *     static DATAITERATOR deleteEntry(T *element){return deleteEntry(element->getId());}
  *     static DATAITERATOR deleteEntry(const unsigned int); //Deletes the entry and will call deletion for associated entries
  *     static SELITERATOR deleteEntry(SELITERATOR iter); //Same but gives the next iterator position
  *     static T *getById(const unsigned int targetId); //Gets item by id. Returns NULL if it could not be found
  *     static const unsigned int &add(T *newElement);
  *     static void setConstraints(int constraints, va_list argptr)
  *     static unsigned int enforceRelation(dependency d); //Each entry which doesn't have at least one first tier relation to d will be delted (total number returned)
  *
  *     static DATAITERATOR begin(ELEMENTTREE &tree=elementTree())
  *     static DATAITERATOR end(ELEMENTTREE &tree=elementTree())
  *     static DATAITERATOR rear(ELEMENTTREE &tree=elementTree())
  *     static DATAITERATOR find(const T val, ELEMENTTREE &tree=elementTree())
  *     static DATAITERATOR findID(unsigned int id, ELEMENTTREE &tree=elementTree())
  *     static DATAITERATOR bound(int lower, int upper, ELEMENTTREE &tree=elementTree())
  *
  *     static SELITERATOR beginSel(ELEMENTTREE &tree=elementTree())
  *     static SELITERATOR endSel(ELEMENTTREE &tree=elementTree())
  *     static SELITERATOR findSel(const T val, ELEMENTTREE &tree=elementTree())
  *     static SELITERATOR findSelID(unsigned int id, ELEMENTTREE &tree=elementTree())
  *     static SELITERATOR fullSelect(bool full=true, ELEMENTTREE &tree = elementTree()) //Do not perform any kind of reduction
  *
  *     static SELITERATOR reduce(bool(*fnct)(JacQL_Base *const&), bool full=true, ELEMENTTREE &tree=elementTree())
  *     static SELITERATOR order(int(*fnct)(JacQL_Base *const&, JacQL_Base *const&), bool full=true, ELEMENTTREE &tree=elementTree())
  *     static SELITERATOR reduce_and_order(bool(*fnct1)(JacQL_Base *const&), int(*fnct2)(JacQL_Base *const&, JacQL_Base *const&),
  *                                         bool full=true, ELEMENTTREE &tree=elementTree())
  *
  *     //FOR ALL constraints: If a negative number is used the invert of the constraint will be returned
  *     static SELITERATOR constrain(bool full=true, int constraints=0, ...);
  *     static SELITERATOR constrain_and_order(int(*fnct)(JacQL_Base *const&, JacQL_Base *const&), bool full=true, int constraints=0, ...);
  *     static T *getByConstrain(bool full=true, int constraints=0, ...);
  *
  *     static SELITERATOR constrain(ELEMENTTREE &tree, bool full=true, int constraints=0, ...);
  *     static SELITERATOR constrain_and_order(ELEMENTTREE &tree, int(*fnct)(JacQL_Base *const&, JacQL_Base *const&),
  *                                            bool full=true, int constraints=0, ...);
  *     static T *getByConstrain(ELEMENTTREE &tree, bool full=true, int constraints=0, ...);
  *
  *     static SELITERATOR query(const char *text, bool full=true, ELEMENTTREE &tree=elementTree());
  *     static SELITERATOR orderQuery(const char *text, bool full=true, ELEMENTTREE &tree=elementTree());
  *     static SELITERATOR query_and_order(const char *que, const char *ord, bool full=true, ELEMENTTREE &tree=elementTree());
  *     static T *getByQuery(const wchar_t *text, bool full=true, ELEMENTTREE &tree = elementTree())
  *
  *     static unsigned int tally(){return elementTree().size();} //Gives the number of current elements in the tree
  *     static unsigned int getAsTally(int constraints=0, ...);
  **/

#define SAVE_CHUNKS     10 //How many entries per file

//Query stuff
#define QUERY_BUFFER_SIZE   2056

#define BEGINQUERY static QueryType type(const wchar_t *text, int *identifier){
#define MIDQUERY return QUERY_INVALID;} double tempDataDouble; long tempDataLong; virtual const void *getData(int identifier){switch (identifier){
#define ENDQUERY default: return NULL;}}

#define ADDQUERYTYPE(id, str, type)     if(wcscmp(text, (str)) == 0) { *identifier = (id); return (type); }

#define RETURNQUERY_STRING(id, el)      case (id) : return (void *)((el));
#define RETURNQUERY_UNICODE(id, el)     case (id) : return (void *)((el));
#define RETURNQUERY_NUMBER(id, el)      case (id) : tempDataLong = (el); return (void *)(&tempDataLong);
#define RETURNQUERY_FLOAT(id, el)       case (id) : tempDataDouble = (el); return (void *)(&tempDataDouble);
#define RETURNQUERY_BOOL(id, el)        case (id) : return (void *)(&(el));

//Casting stuff
#define INPUT_SINGLE                        (JacQL_Base *const& tempJacQL_BaseCast)
#define CASTINPUT_SINGLE(type, el)          type const el = dynamic_cast<type const>((tempJacQL_BaseCast));
#define INPUT_DOUBLE                        (JacQL_Base *const& tempJacQL_BaseCast1, JacQL_Base *const& tempJacQL_BaseCast2)
#define CASTINPUT_DOUBLE(type, el1, el2)    type const el1 = dynamic_cast<type const>((tempJacQL_BaseCast1));type const el2 = dynamic_cast<type const>((tempJacQL_BaseCast2));

//Iterator Helper
#define DATA_ITER(type)                         typename JacQL_Template<type>::dataIterator
#define SEL_ITER(type)                          typename JacQL_Template<type>::selIterator
#define FIRST_TIER_ITER(type, name, dependency) typename JacQL_Template<type>::dataIterator name = type::begin(fTR(dependency)); name != type::end(fTR(dependency)); name++

//Getters and Setters (Called Secure Variables). Place in public space!
#define SEC_VAR(type, name)     private: type name; public: type get_##name() const {return name;} void set_##name(type temp, const bool secure=true){name=temp;if(secure){saveRequest();}}
#define SEC_VAR_STRING(name)    private: char* name; public: const char* get_##name() const {return name;} void set_##name(const char* temp, const bool secure=true){int length = strlen(temp)+1; if(name != NULL){delete[] name;} name = new char[length]; strcpy(name, temp); if(secure){saveRequest();}}
#define SEC_VAR_UNICODE(name)   private: wchar_t* name; public: const wchar_t* get_##name() const {return name;} void set_##name(const wchar_t* temp, const bool secure=true){int length = wcslen(temp)+1; if(name != NULL){delete[] name;} name = new wchar_t[length]; wcscpy(name, temp); if(secure){saveRequest();}}

#include "InvalidFileException.h"
#include "JacQL_Base.h"
#include <cstdlib>
#include <cstdarg>
using namespace std;


//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * Exception * * * /////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

class JacQLException_Template : public JacQLException
{
    public:
        JacQLException_Template(const char *desc) : JacQLException(desc) {;}
};

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * Template * * * //////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

template <class T>
class JacQL_Template : public JacQL_Base
{
    typedef AVLtree<JacQL_Base *> ELEMENTTREE;
    public:

        JacQL_Template<T>(){saveRequest();deletionFlag=false;}
        virtual ~JacQL_Template<T>(){saveRequest();}

        virtual void setDependency(dependency, unsigned int); //Sets the dependency to another entry

        virtual T& operator=(const T&)=0;

        //Standard Iterator
        class dataIterator: public ELEMENTTREE::treeIterator
        {
            public:
                dataIterator(const ELEMENTTREE::treeIterator& other) : ELEMENTTREE::treeIterator(other) {;}
                T *operator*() const
                {
                    return static_cast<T *>(ELEMENTTREE::treeIterator::operator*());
                }
                T **operator->() const {return &(operator*());}
        };
        typedef JacQL_Template<T>::dataIterator DATAITERATOR;

        //Select Iterator
        class selIterator: public ELEMENTTREE::selectIterator
        {
            public:
                selIterator(const ELEMENTTREE::selectIterator& other) : ELEMENTTREE::selectIterator(other) {;}
                T *operator*() const
                {
                    return static_cast<T *>(ELEMENTTREE::selectIterator::operator*());
                }
                T **operator->() const {return &(operator*());}
        };
        typedef JacQL_Template<T>::selIterator SELITERATOR;

        static void setup(dependency, bool=true); //Sets the dependencyTruthTable to all false and sets the own Entry and if it is a weak Table
        static void loadAll(const char *, dependency) throw (InvalidFileException); //Loads all values in the given path
        static void saveAll(const char *, dependency) throw (InvalidFileException); //Saves all values in the given path
        static void deleteEntriesWithDependency(dependency, unsigned int); //Deletes all entries with dependency to the stated id and calls deletion recursively
        static void updateEntriesWithDependency(dependency, unsigned int, unsigned int); //Changes the dependencies of all entries with this one
        static DATAITERATOR deleteEntry(T *element){return deleteEntry(element->getId());}
        static DATAITERATOR deleteEntry(const unsigned int); //Deletes the entry and will call deletion for associated entries
        static SELITERATOR deleteEntry(SELITERATOR iter); //Same but gives the next iterator position
        static T *getById(const unsigned int targetId); //Gets item by id. Returns NULL if it could not be found
        static const unsigned int &add(T *newElement);
        static void setConstraints(int constraints, va_list argptr);
        static unsigned int enforceRelation(dependency d);

        static DATAITERATOR begin(ELEMENTTREE &tree = elementTree()) {return tree.begin();}
        static DATAITERATOR end(ELEMENTTREE &tree = elementTree()) {return tree.end();}
        static DATAITERATOR rear(ELEMENTTREE &tree = elementTree()) {return tree.rear();}
        static DATAITERATOR find(const T val, ELEMENTTREE &tree = elementTree()) {return tree.find(val);}
        static DATAITERATOR findID(unsigned int id, ELEMENTTREE &tree = elementTree()) {return tree.findID(id);}
        static DATAITERATOR bound(int lower, int upper, ELEMENTTREE &tree = elementTree()) {return tree.bound(lower, upper);}

        static SELITERATOR beginSel(ELEMENTTREE &tree = elementTree()) {return tree.beginSelection();}
        static SELITERATOR endSel(ELEMENTTREE &tree = elementTree()) {return tree.endSelection();}
        static SELITERATOR findSel(const T val, ELEMENTTREE &tree = elementTree()) {return tree.findSelection(val);}
        static SELITERATOR findSelID(unsigned int id, ELEMENTTREE &tree = elementTree()) {return tree.findSelectionID(id);}
        static SELITERATOR fullSelect(bool full=true, ELEMENTTREE &tree = elementTree()) {tree.reduce(JacQL_Base::fullSelect, full);return beginSel(tree);}

        static SELITERATOR reduce(bool(*fnct)(JacQL_Base *const&), bool full=true, ELEMENTTREE &tree = elementTree()){tree.reduce(fnct, full);return beginSel(tree);}
        static SELITERATOR order(int(*fnct)(JacQL_Base *const&, JacQL_Base *const&), bool full=true, ELEMENTTREE &tree = elementTree()){tree.order(fnct, full);return beginSel(tree);}
        static SELITERATOR reduce_and_order(bool(*fnct1)(JacQL_Base *const&), int(*fnct2)(JacQL_Base *const&, JacQL_Base *const&), bool full=true, ELEMENTTREE &tree = elementTree())
            {tree.reduce_and_order(fnct1, fnct2, full);return beginSel(tree);}

        //FOR ALL constraints: If a negative number is used the invert of the constraint will be returned
        static SELITERATOR constrain(bool full=true, int constraints=0, ...);
        static SELITERATOR constrain_and_order(int(*fnct)(JacQL_Base *const&, JacQL_Base *const&), bool full=true, int constraints=0, ...);
        static T *getByConstrain(bool full=true, int constraints=0, ...);

        //External Element Tree
        static SELITERATOR constrain(ELEMENTTREE &tree, bool full=true, int constraints=0, ...);
        static SELITERATOR constrain_and_order(ELEMENTTREE &tree, int(*fnct)(JacQL_Base *const&, JacQL_Base *const&), bool full=true, int constraints=0, ...);
        static T *getByConstrain(ELEMENTTREE &tree, bool full=true, int constraints=0, ...);

        static SELITERATOR query(const wchar_t *text, bool full=true, ELEMENTTREE &tree = elementTree());
        static SELITERATOR orderQuery(const wchar_t *text, bool full=true, ELEMENTTREE &tree = elementTree());
        static SELITERATOR query_and_order(const wchar_t *que, const wchar_t *ord, bool full=true, ELEMENTTREE &tree = elementTree());
        static T *getByQuery(const wchar_t *text, bool full=true, ELEMENTTREE &tree = elementTree());

        static unsigned int tally(){return elementTree().size();} //Gives the number of current elements in the tree
        static unsigned int getAsTally(int constraints=0, ...);

        static dependency getOwnEntry(){return ownEntry;}
        static dependency firstTierAssociator[D_ENDTAG];

    protected:
        //**** AVOID Static Initialization Order Fiasco! Therefore: Construct On First Use Idiom *************//
        static ELEMENTTREE &elementTree() //Needs to be accessible from below for load() and save()
        {
            static ELEMENTTREE *tree = new AVLtree<JacQL_Base *>(JacQL_Base::compare, JacQL_Base::compareID);
            return *tree;
        }

        static dependency ownEntry; //Own dependency

        //Place a save request into the AVL Tree -if changes have occurred-
        virtual void saveRequest(){requestTree().insert(getId()/SAVE_CHUNKS);}
        static void saveRequest(unsigned int id){requestTree().insert(id/SAVE_CHUNKS);}

        virtual const Constrainer& getConstraints(){return constraintStruct;};

        /** Data types:
         *  0 Invalid
         *  1 ASCII String
         *  2 Unicode String
         *  3 Integer(Long)
         *  4 Float(Double)
         *  5 Boolean
         **/
         /** Comparisons:
          *  0x01  ==
          *  0x02  <
          *  0x04  >
          *  0x08  CONTAINS
          *  0x10  !CONTAINS
          **/
          /** Comparisons Mask (queryComparison):
          *  0x01  ==
          *  0x02  <
          *  0x03  <=
          *  0x04  >
          *  0x05  >=
          *  0x06  !=
          *  0x08  CONTAINS
          *  0x10  !CONTAINS
          **/
        //Query stuff
        virtual const QueryType& getQueryType(){return queryType;}
        virtual const int& getQueryId(){return queryId;}
        virtual const void *getQueryElement(){return queryElement;}
        virtual const int& getQueryComparison(){return queryComparison;}
        virtual const QueryType& getCompareQueryType(){return compareQueryType;}
        virtual const int& getCompareQueryId(){return compareQueryId;}
        virtual const bool getCompareQueryDesc(){return compareQueryDesc;}

    private:
        static Constrainer constraintStruct;

        static AVLtree<unsigned int> &requestTree()
        {
            static AVLtree<unsigned int> *tree = new AVLtree<unsigned int>(JacQL_Base::compareUI, JacQL_Base::compareUI);
            return *tree;
        }
        static MinHeap<unsigned int> &idPool() //Available ids within the range of begin to rear
        {
            static MinHeap<unsigned int> *heap = new MinHeap<unsigned int>(JacQL_Base::compareUI);
            return *heap;
        }

        //Query stuff
        static QueryType queryType;

        static int queryId;
        static void *queryElement;
        static int queryComparison;
        static QueryType compareQueryType;
        static int compareQueryId;
        static bool compareQueryDesc;

        static void setupOrderQuery(const wchar_t *text);
        static void setupQuery(const wchar_t *text);


        static bool dependencyTruthTable[D_ENDTAG];
        static bool weakTable; //Weak tables don't cause a cascade of deletion. Instead all dependent entries will be
                    //updated to id 0. Therefore no further deletions will be triggered but a clean exit ensured
        static unsigned int getNextAvailableId();
        static void clearList(); //Deletes all entries in the list, BUT leaves dependencies alone

        static unsigned int loadBound(const char *path, unsigned int boundIdentifier) throw (InvalidFileException);
        static unsigned int saveBound(const char *path, unsigned int boundIdentifier) throw (InvalidFileException);

        bool deletionFlag; //Set this flag in deleteEntry which marks that this entry is about to be deleted
            //this way we avoid a circular deletion chain
};


template<class T>
dependency JacQL_Template<T>::firstTierAssociator[];

template<class T>
dependency JacQL_Template<T>::ownEntry;

template<class T>
Constrainer JacQL_Template<T>::constraintStruct;

template<class T>
bool JacQL_Template<T>::dependencyTruthTable[];

template<class T>
bool JacQL_Template<T>::weakTable;

template<class T>
QueryType JacQL_Template<T>::queryType;

template<class T>
int JacQL_Template<T>::queryId;

template<class T>
void *JacQL_Template<T>::queryElement;

template<class T>
int JacQL_Template<T>::queryComparison;

template<class T>
QueryType JacQL_Template<T>::compareQueryType;

template<class T>
int JacQL_Template<T>::compareQueryId;

template<class T>
bool JacQL_Template<T>::compareQueryDesc;


template<class T>
void JacQL_Template<T>::setDependency(dependency d, unsigned int d_id)
{
    if(d==ownEntry) //Not allowed to depend on itself
        throw JacQLException_Template("JacQLException: Entry is not allowed to depend on an element of the same class!");
    //First check the truth table
    if(dependencyTruthTable[d]==false && d_id!=0)
        dependencyTruthTable[d]=true;
    JacQL_Base::setDependency(d, d_id);
    saveRequest();
}

template<class T>
void JacQL_Template<T>::setup(dependency d, bool weak)
{
    for(int i=0; i<D_ENDTAG; i++)
    {
        dependencyTruthTable[i] = false;
        firstTierAssociator[i] = D_ENDTAG;
    }
    ownEntry = d;
    weakTable = weak;
    queryType = QUERY_INVALID;
}

template<class T>
void JacQL_Template<T>::loadAll(const char *path, dependency d) throw(InvalidFileException)
{
    unsigned int fileId;
    char fileName[strlen(path)+5];
    ios::openmode mode = ios::in | ios::binary;
    if(d!=D_ENDTAG && d!=ownEntry) //Not for this entry
        return;
    sprintf(fileName, "%s.dat", path);
    ifstream inStream(fileName, mode);
    if(!inStream)
        throw InvalidFileException(fileName, ownEntry);
    //Clear the elementList here
    for(DATAITERATOR pos=begin(); pos!=end(); )
        pos = deleteEntry(*pos); //Also deletes other entries in DB, ok as we can assume that other
            //elements will be loaded in right order
    inStream.read((char *)&fileId, sizeof(unsigned int));
    while(!inStream.eof()) //EOF only set when tried to read at file end
    {
        loadBound(path, fileId);
        //And attempt next entry
        inStream.read((char *)&fileId, sizeof(unsigned int));
    }
    inStream.close();

    //Goes over all elements and fills the idPool according to holes
    unsigned int i = 1U;
    for(DATAITERATOR iter = begin(); iter != end(); i++)
    {
        if((*iter)->getId() != i)
            idPool().cleanPush(i);
        else
            iter++;
    }
}

template<class T>
unsigned int JacQL_Template<T>::loadBound(const char *path, const unsigned int boundIdentifier) throw(InvalidFileException)
{
    T *tempEntry;
    unsigned int tempId;
    unsigned int counter = 0U;
    char fileName[strlen(path)+15];
    ios::openmode mode = ios::in | ios::binary;
    sprintf(fileName, "%s%u.dat", path, boundIdentifier);
    ifstream inStream(fileName, mode);
    if(!inStream)
        throw InvalidFileException(fileName, ownEntry);
    inStream.read((char *)&tempId, sizeof(unsigned int));
    while(!inStream.eof()) //EOF only set when tried to read at file end
    {
        counter++;
        if(tempId/SAVE_CHUNKS != boundIdentifier)
            throw JacQLException_Template("Fatal JacQLException: Found an out of range id in file while loading!");
        tempEntry = new T();
        tempEntry->id = tempId;
        for(int i=0; i<D_ENDTAG; i++)
        {
            inStream.read((char *)&(tempEntry->dependencies[i]), sizeof(unsigned int));
            if(tempEntry->dependencies[i]!=0)
                dependencyTruthTable[i] = true;
        }
        tempEntry->load(inStream);
        if(!elementTree().insert(tempEntry))
            throw JacQLException("JacQLException: Hit illegal duplicate entry while loading JacQL database!");
        //And attempt next entry
        inStream.read((char *)&tempId, sizeof(unsigned int));
    }
    inStream.close();
    return counter;
}

template<class T>
void JacQL_Template<T>::saveAll(const char *path, dependency d) throw(InvalidFileException)
{
    if(d!=D_ENDTAG && d!=ownEntry) //Not for this entry
        return;

    MinHeap<unsigned int> boundHeap(compareUI);
    unsigned int fileId;
    char fileName[strlen(path)+15];
    ios::openmode mode = ios::in | ios::binary;
    if(d!=D_ENDTAG && d!=ownEntry) //Not for this entry
        return;

    sprintf(fileName, "%s.dat", path);
    ifstream inStream(fileName, mode);
    if(!inStream) //File not found? Just delete files and save everything afterwards
    {
        ELEMENTTREE::treeIterator rear = elementTree().rear();
        int topChunck;
        if(rear != elementTree().end())
            topChunck = (*rear)->getId()/SAVE_CHUNKS;
        else
            topChunck = -1;
        clearUpDir(path);
        mode = ios::out | ios::trunc | ios::binary;
        ofstream outStream(fileName, mode);
        if(!outStream)
            throw InvalidFileException(fileName, ownEntry);
        for(int i=0; i<=topChunck; i++)
        {
            if(saveBound(path, i)) //Only if actually saved something to the file we keep it alive
                outStream.write((char *)&i, sizeof(unsigned int));
        }
        requestTree().clear();
        outStream.close();
        return;
    }
    inStream.read((char *)&fileId, sizeof(unsigned int));
    while(!inStream.eof()) //EOF only set when tried to read over the end
    {
        boundHeap.cleanPush(fileId);
        inStream.read((char *)&fileId, sizeof(unsigned int));//And attempt next entry
    }
    inStream.close();

    mode = ios::out | ios::trunc | ios::binary;
    ofstream outStream(fileName, mode);
    if(!outStream)
        throw InvalidFileException(fileName, ownEntry);

    const unsigned int INFINITY = 0UL - 1UL;
    AVLtree<unsigned int>::treeIterator iter = requestTree().begin();
    unsigned int request = (iter!=requestTree().end())?*iter:INFINITY;
    unsigned int file = (!boundHeap.empty())?boundHeap.cleanPull():INFINITY;
    while(request != INFINITY || file != INFINITY)
    {
        if(request > file)
        {
            outStream.write((char *)&file, sizeof(unsigned int));
            file = (!boundHeap.empty())?boundHeap.cleanPull():INFINITY;
        }
        else
        {
            if(saveBound(path, request)) //Only if actually saved something to the file we keep it alive
                outStream.write((char *)&request, sizeof(unsigned int));
            else //No entries saved? Delete file instead
            {
                sprintf(fileName, "%s%u.dat", path, request);
                remove(fileName);
            }
            if(request == file)
                file = (!boundHeap.empty())?boundHeap.cleanPull():INFINITY;
            iter++;
            request = (iter!=requestTree().end())?*iter:INFINITY;
        }
    }
    requestTree().clear();
    outStream.close();
}

template<class T>
unsigned int JacQL_Template<T>::saveBound(const char *path, const unsigned int boundIdentifier) throw(InvalidFileException)
{
    unsigned int counter = 0U;
    char fileName[strlen(path)+15];
    ios::openmode mode = ios::out | ios::trunc | ios::binary;
    sprintf(fileName, "%s%u.dat", path, boundIdentifier);
    ofstream outStream(fileName, mode);
    if(!outStream)
        throw InvalidFileException(path, ownEntry);
    for(DATAITERATOR pos=bound(boundIdentifier*SAVE_CHUNKS, (boundIdentifier+1)*SAVE_CHUNKS-1);
        pos!=end(); pos++)
    {
        counter++;
        outStream.write((char *)&((*pos)->id), sizeof(unsigned int));
        for(int i=0; i<D_ENDTAG; i++)
            outStream.write((char *)&((*pos)->dependencies[i]), sizeof(unsigned int));
        (*pos)->save(outStream);
    }
    outStream.close();
    return counter;
}

template<class T>
void JacQL_Template<T>::deleteEntriesWithDependency(dependency d, unsigned int d_id)
{
    if(d==ownEntry || dependencyTruthTable[d]==false) //We called ourselves or we have nothing to do with this one
        return;
    for(DATAITERATOR pos=begin(); pos!=end(); )
    {
        if((*pos)->dependencies[d]==d_id) //We found one
            pos = deleteEntry(*pos);
        else //We haven't found anything. Go on!
            pos++;
    }
}

template<class T>
void JacQL_Template<T>::updateEntriesWithDependency(dependency d, unsigned int d_old_id, unsigned int d_new_id)
{
    if(d==ownEntry || dependencyTruthTable[d]==false) //We called ourselves or we have nothing to do with this one
        return;
    for(DATAITERATOR pos = begin(); pos != end(); pos++)
    {
        if((*pos)->dependencies[d]==d_old_id)
            (*pos)->setDependency(d, d_new_id);
    }
}

template<class T>
typename JacQL_Template<T>::dataIterator JacQL_Template<T>::deleteEntry(const unsigned int targetId)
{
    DATAITERATOR pos = elementTree().findID(targetId);
    if((*pos)->deletionFlag) //Already about to be deleted
        return ++pos;
    (*pos)->deletionFlag = true;
    if(!weakTable)
        deleteEntryInDatabase(ownEntry, targetId);
    else
        updateEntryInDatabase(ownEntry, targetId, 0);
    pos = elementTree().deleteKeyID(targetId);
    idPool().cleanPush(targetId);
    saveRequest(targetId);
    return pos;
}

template<class T>
typename JacQL_Template<T>::selIterator JacQL_Template<T>::deleteEntry(JacQL_Template<T>::selIterator iter)
{
    unsigned int targetId = (*iter)->getId();
    if(!weakTable)
        deleteEntryInDatabase(ownEntry, targetId);
    else
        updateEntryInDatabase(ownEntry, targetId, 0);
    SELITERATOR pos = elementTree().deleteKey(iter);
    idPool().cleanPush(targetId);
    saveRequest(targetId);
    return pos;
}

template<class T>
T *JacQL_Template<T>::getById(const unsigned int targetId)
{
    try{
        return static_cast<T *>(elementTree()[targetId]);
    } catch (JacQLException_AVLtree) {
        return NULL;
    }
}

template<class T>
const unsigned int &JacQL_Template<T>::add(T *newElement)
{
    newElement->id = getNextAvailableId();
    elementTree().insert(newElement);
    saveRequest(newElement->id);
    return newElement->id;
}

template<class T>
void JacQL_Template<T>::setConstraints(int constraints, va_list argptr)
{
    constraintStruct.num = constraints;

    if(constraints<0)
        constraints *= -1;

    if(constraintStruct.conTable != NULL)
    {
        delete[] constraintStruct.conTable;
        delete[] constraintStruct.conId;
    }

    constraintStruct.conTable = new dependency[constraints];
    constraintStruct.conId = new unsigned int[constraints];

    for(int i=0; i<constraints; i++)
    {
        constraintStruct.conTable[i] = (dependency)(va_arg(argptr, int));
        constraintStruct.conId[i] = va_arg(argptr, unsigned int);
    }
    va_end(argptr);
}

template<class T>
unsigned int JacQL_Template<T>::enforceRelation(dependency d)
{
    unsigned int counter = 0U;
    if(firstTierAssociator[d] == D_ENDTAG)
        throw JacQLException_Template("Fatal JacQLException: Enforced relation is not coupled with an association!");
    for(DATAITERATOR iter = begin(); iter!=end(); )
    {
        if((*iter)->firstTierRelations[d].empty())
        {
            counter++;
            iter = deleteEntry(*iter);
        }
        else
            iter++;
    }
    return counter;
}

template<class T>
typename JacQL_Template<T>::selIterator JacQL_Template<T>::constrain(bool full, int constraints, ...)
{
    va_list argptr;
    va_start(argptr, constraints);
    setConstraints(constraints, argptr);
    elementTree().reduce(JacQL_Base::applyConstraints, full);
    return elementTree().beginSelection();
}

template<class T>
typename JacQL_Template<T>::selIterator JacQL_Template<T>::constrain_and_order(int(*fnct)(JacQL_Base *const&, JacQL_Base *const&), bool full, int constraints, ...)
{
    va_list argptr;
    va_start(argptr, constraints);
    setConstraints(constraints, argptr);
    elementTree().reduce_and_order(JacQL_Base::applyConstraints, fnct, full);
    return elementTree().beginSelection();
}

template<class T>
T *JacQL_Template<T>::getByConstrain(bool full, int constraints, ...)
{
    T *ret;
    va_list argptr;
    va_start (argptr, constraints);
    setConstraints(constraints, argptr);
    try
    {
        ret = static_cast<T *>(elementTree().lookup(JacQL_Base::applyConstraints, full));
    }
    catch(JacQLException_AVLtree e)
    {
        return NULL;
    }
    return ret;
}


template<class T>
typename JacQL_Template<T>::selIterator JacQL_Template<T>::constrain(AVLtree<JacQL_Base *> &tree, bool full, int constraints, ...)
{
    va_list argptr;
    va_start(argptr, constraints);
    setConstraints(constraints, argptr);
    tree.reduce(JacQL_Base::applyConstraints, full);
    return tree.beginSelection();
}

template<class T>
typename JacQL_Template<T>::selIterator JacQL_Template<T>::constrain_and_order(AVLtree<JacQL_Base *> &tree, int(*fnct)(JacQL_Base *const&, JacQL_Base *const&), bool full, int constraints, ...)
{
    va_list argptr;
    va_start(argptr, constraints);
    setConstraints(constraints, argptr);
    tree.reduce_and_order(JacQL_Base::applyConstraints, fnct, full);
    return tree.beginSelection();
}

template<class T>
T *JacQL_Template<T>::getByConstrain(ELEMENTTREE &tree, bool full, int constraints, ...)
{
    T *ret;
    va_list argptr;
    va_start (argptr, constraints);
    setConstraints(constraints, argptr);
    try
    {
        ret = static_cast<T *>(tree.lookup(JacQL_Base::applyConstraints, full));
    }
    catch(JacQLException_AVLtree e)
    {
        return NULL;
    }
    return ret;
}

template<class T>
typename JacQL_Template<T>::selIterator JacQL_Template<T>::query(const wchar_t *text, bool full, AVLtree<JacQL_Base *> &tree)
{
    setupQuery(text);
    tree.reduce(JacQL_Base::applyQuery, full);
    return tree.beginSelection();
}

template<class T>
void JacQL_Template<T>::setupQuery(const wchar_t *text)
{
    wchar_t first[QUERY_BUFFER_SIZE], second[QUERY_BUFFER_SIZE], third[QUERY_BUFFER_SIZE];
    int fetched = swscanf(text, L"%ls %ls %ls", first, second, third);
    if(fetched == 2)
        third[0] = L'\0';
    else if(fetched != 3)
        throw JacQLException_Template("JacQLException: Could not fetch all three strings for query!");

    switch (queryType) //DELETE OLD ELEMENT
    {
        case QUERY_STRING: //String
            delete [] (char *)(queryElement);
            break;
        case QUERY_UNICODE: //Unicode String
            delete [] (wchar_t *)(queryElement);
            break;
        case QUERY_NUMBER: //Long
            delete (long *)(queryElement);
            break;
        case QUERY_FLOAT: //Double
            delete (double *)(queryElement);
            break;
        case QUERY_BOOL: //Boolean
            delete (bool *)(queryElement);
        default: //No previous query
            break;
    }

    queryType = T::type(first, &queryId);
    if(queryType == QUERY_INVALID)
        throw JacQLException_Template("JacQLException: Query element not found!");

    if(wcscmp(second, L"==") == 0)
        queryComparison = 0x01;
    else if(wcscmp(second, L"<") == 0)
        queryComparison = 0x02;
    else if(wcscmp(second, L"<=") == 0)
        queryComparison = 0x03;
    else if(wcscmp(second, L">") == 0)
        queryComparison = 0x04;
    else if(wcscmp(second, L">=") == 0)
        queryComparison = 0x05;
    else if(wcscmp(second, L"!=") == 0)
        queryComparison = 0x06;
    else if(wcscmp(second, L"CONTAINS") == 0)
        queryComparison = 0x08;
    else if(wcscmp(second, L"!CONTAINS") == 0)
        queryComparison = 0x10;
    else
        throw JacQLException_Template("JacQLException: Unable to fetch comparison operator!");

    switch (queryType)
    {
        case QUERY_STRING: //String
            queryElement = new char[wcslen(third)+1];
            wcstombs((char *)queryElement, third, wcslen(third)+1);
            break;
        case QUERY_UNICODE: //Unicode String
            queryElement = new wchar_t[wcslen(third)+1];
            wcscpy((wchar_t *) queryElement, third);
            break;
        case QUERY_NUMBER: //Long
            queryElement = new long;
            if(!swscanf(third, L"%ld", (long *)queryElement))
                throw JacQLException_Template("JacQLException: Unable to fetch long integer from query!");
            break;
        case QUERY_FLOAT: //Double
            queryElement = new double;
            if(!swscanf(third, L"%f", (double *)queryElement))
                throw JacQLException_Template("JacQLException: Unable to fetch double from query!");
            break;
        case QUERY_BOOL: //Boolean
            {
                queryElement = new bool;

                bool *element = (bool *)(queryElement);
                if(wcscmp(third, L"TRUE") == 0)
                    *element = true;
                else if(wcscmp(third, L"FALSE") == 0)
                    *element = false;
                else
                    throw JacQLException_Template("JacQLException: Unable to fetch boolean value from query!");
                break;
                }
        default:
            throw JacQLException_Template("Fatal JacQLException: Query Type is out of range!");
    }
}

template<class T>
typename JacQL_Template<T>::selIterator JacQL_Template<T>::orderQuery(const wchar_t *text, bool full, AVLtree<JacQL_Base *> &tree)
{
    setupOrderQuery(text);
    tree.order(JacQL_Base::applyCompareQuery, full);
    return tree.beginSelection();
}

template<class T>
void JacQL_Template<T>::setupOrderQuery(const wchar_t *text)
{
    wchar_t first[QUERY_BUFFER_SIZE], second[QUERY_BUFFER_SIZE];
    if(swscanf(text, L"%ls %ls", first, second) != 2)
        throw JacQLException_Template("JacQLException: Could not fetch all two strings for query!");

    compareQueryType = T::type(second, &compareQueryId);
    if(compareQueryType == QUERY_INVALID)
        throw JacQLException_Template("JacQLException: Query element not found!");

    if(wcscmp(first, L"ASC") == 0) //Standard Ascending
        compareQueryDesc = false;
    else if(wcscmp(first, L"DESC") == 0)
        compareQueryDesc = true;
    else
        throw JacQLException_Template("JacQLException: Order keyword is not ASC or DESC!");
}

template<class T>
typename JacQL_Template<T>::selIterator JacQL_Template<T>::query_and_order(const wchar_t *que, const wchar_t *ord, bool full, AVLtree<JacQL_Base *> &tree)
{
    setupQuery(que);
    setupOrderQuery(ord);
    tree.reduce_and_order(JacQL_Base::applyQuery, JacQL_Base::applyCompareQuery, full);
    return tree.beginSelection();
}

template<class T>
T *JacQL_Template<T>::getByQuery(const wchar_t *text, bool full, ELEMENTTREE &tree)
{
    T *ret;
    setupQuery(text);
    try
    {
        ret = static_cast<T *>(tree.lookup(JacQL_Base::applyQuery, full));
    }
    catch(JacQLException_AVLtree e)
    {
        return NULL;
    }
    return ret;
}

template<class T>
unsigned int JacQL_Template<T>::getAsTally(int constraints, ...)
{
    va_list argptr;
    va_start(argptr, constraints);
    setConstraints(constraints, argptr);
    return elementTree().tally(JacQL_Base::applyConstraints);
}

template<class T>
unsigned int JacQL_Template<T>::getNextAvailableId()
{
    if(!idPool().empty())
        return idPool().cleanPull();
    else
    {
        ELEMENTTREE::treeIterator rear = elementTree().rear();
        if(rear != elementTree().end())
            return (*rear)->getId() + 1;
    }
    return 1U;
}



#endif // JACQL_TEMPLATE_H
